@ Some functions and tests to increase performance in drawing.cpp and custom.cpp

.arm

.global NEON_shift32plus
.global NEON_shift64plus
.global ARM_ReadCustomchipLong
.global ARM_doline_n1
.global NEON_doline_n2
.global NEON_doline_n3
.global NEON_doline_n4
.global NEON_doline_n6
.global NEON_doline_n8


.text

.align 8

@----------------------------------------------------------------
@ NEON_shift32plus
@
@ r0: uae_u32   *p
@ r1: int       n
@
@ void NEON_shift32plus(uae_u32 *p, int n);
@
@----------------------------------------------------------------
NEON_shift32plus:
  vldr.64   d0, [r0]
  vmov.32   d1[0], r1
  vshl.U64  d0, d0, d1
  vstr.64   d0, [r0]
  bx        lr


.align 8

@----------------------------------------------------------------
@ NEON_shift64plus
@
@ r0: uae_u32   *p
@ r1: int       n
@
@ void NEON_shift64plus(uae_u32 *p, int n);
@
@----------------------------------------------------------------
NEON_shift64plus:
  vldr.64   d1, [r0]
  vldr.64   d0, [r0, #8]
  vmov.u8   d2, #0
  vmov.32   d3[0], r1
  vshl.U64  q0, q0, q1
  vstr.64   d1, [r0]
  vstr.64   d0, [r0, #8]
  bx        lr


.align 8

@----------------------------------------------------------------
@ NEON_ReadCustomchipLong
@
@ r0: uae_u32   *p
@
@ uae_u32 NEON_ReadCustomchipLong(uae_u32 *p);
@
@----------------------------------------------------------------
ARM_ReadCustomchipLong:
  ldr       r0, [r0]
  rev       r0, r0
  rev16     r0, r0
  bx        lr


.align 8

@----------------------------------------------------------------
@ ARM_doline_n1
@
@ r0: uae_u32   *pixels
@ r1: int       wordcount
@ r2: int       lineno 
@
@ void ARM_doline_n1(uae_u32 *pixels, int wordcount, int lineno);
@
@----------------------------------------------------------------
ARM_doline_n1:
  stmdb     sp!, {r4-r9, lr}

  mov       r3, #1600
  mul       r2, r2, r3
  ldr       r3, =line_data
  add       r3, r3, r2           @ real_bplpt[0]

  ldr       r4, =Lookup_doline_n1

  sub       r3, r3, #4
  
ARM_doline_n1_loop:
  ldr       r2, [r3, #4]!
@  add       r3, r3, #4

  ubfx      r5, r2, #28, #4
  add       r5, r4, r5, lsl #2
  ldr       r6, [r5]
  
  ubfx      r5, r2, #24, #4
  add       r5, r4, r5, lsl #2
  ldr       r7, [r5]

  ubfx      r5, r2, #20, #4
  add       r5, r4, r5, lsl #2
  ldr       r8, [r5]

  ubfx      r5, r2, #16, #4
  add       r5, r4, r5, lsl #2
  ldr       r9, [r5]
  stmia     r0!, {r6-r9}
  
  ubfx      r5, r2, #12, #4
  add       r5, r4, r5, lsl #2
  ldr       r6, [r5]

  ubfx      r5, r2, #8, #4
  add       r5, r4, r5, lsl #2
  ldr       r7, [r5]

  ubfx      r5, r2, #4, #4
  add       r5, r4, r5, lsl #2
  ldr       r8, [r5]

  ubfx      r5, r2, #0, #4
  add       r5, r4, r5, lsl #2
  ldr       r9, [r5]
  stmia     r0!, {r6-r9}

  subs      r1, r1, #1
  bgt       ARM_doline_n1_loop
  
  ldmia     sp!, {r4-r9, pc}


.align 8

@----------------------------------------------------------------
@ NEON_doline_n2
@
@ r0: uae_u32   *pixels
@ r1: int       wordcount
@ r2: int       lineno 
@
@ void NEON_doline_n2(uae_u32 *pixels, int wordcount, int lineno);
@
@----------------------------------------------------------------
NEON_doline_n2:

  mov       r3, #1600
  mul       r2, r2, r3
  ldr       r3, =line_data
  add       r2, r3, r2           @ real_bplpt[0]
  add       r3, r2, #200

@ Load masks to registers
  vmov.u8   d10, #0x55
  vmov.u8   d11, #0x33
  vmov.u8   d12, #0x0f

NEON_doline_n2_loop:
  @ Load data as early as possible
  vldmia    r2!, {d4}
  vldmia    r3!, {d6}

@      MERGE (b6, b7, 0x55555555, 1);
  vshr.u8   d8, d4, #1      @ tmpb = b >> shift
  vshl.u8   d9, d6, #1      @ tmpa = a << shift
  vbit.u8   d6, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d10     @ b = b and bit set from tmpa if mask is false

@      MERGE_0(b4, b6, 0x33333333, 2);
  vshr.u8   d8, d6, #2		  @ tmp = b >> shift
  vand.8    d2, d8, d11     @ a = tmp & mask
  vand.8    d6, d6, d11     @ b = b & mask
@      MERGE_0(b5, b7, 0x33333333, 2);
  vshr.u8   d8, d4, #2		  @ tmp = b >> shift
  vand.8    d0, d8, d11     @ a = tmp & mask
  vand.8    d4, d4, d11     @ b = b & mask

@      MERGE_0(b0, b4, 0x0f0f0f0f, 4);
  vshr.u8   d8, d2, #4		  @ tmp = b >> shift
  vand.8    d3, d8, d12     @ a = tmp & mask
  vand.8    d2, d2, d12     @ b = b & mask
@      MERGE_0(b1, b5, 0x0f0f0f0f, 4);
  vshr.u8   d8, d0, #4			@ tmp = b >> shift
  vand.8    d1, d8, d12     @ a = tmp & mask
  vand.8    d0, d0, d12     @ b = b & mask
@      MERGE_0(b2, b6, 0x0f0f0f0f, 4);
  vshr.u8   d8, d6, #4			@ tmp = b >> shift
  vand.8    d7, d8, d12     @ a = tmp & mask
  vand.8    d6, d6, d12     @ b = b & mask
@      MERGE_0(b3, b7, 0x0f0f0f0f, 4);
  vshr.u8   d8, d4, #4			@ tmp = b >> shift
  vand.8    d5, d8, d12     @ a = tmp & mask
  vand.8    d4, d4, d12     @ b = b & mask
  
  vzip.8    d3, d7
  vzip.8    d1, d5
  vzip.8    d2, d6
  vzip.8    d0, d4

  vzip.8    d3, d1
  vzip.8    d2, d0
  vzip.32   d3, d2
  vzip.32   d1, d0

  vst1.8    {d0, d1, d2, d3}, [r0]!
  
  cmp       r1, #1    @ Exit from here if odd number of words
  bxeq      lr
   
  subs      r1, r1, #2    @ We handle 2 words (64 bit) per loop: wordcount -= 2

  vzip.8    d7, d5
  vzip.8    d6, d4
  vzip.32   d7, d6
  vzip.32   d5, d4

  vst1.8    {d4, d5, d6, d7}, [r0]!

  bgt       NEON_doline_n2_loop
  
NEON_doline_n2_exit:
  bx        lr


.align 8

@----------------------------------------------------------------
@ NEON_doline_n3
@
@ r0: uae_u32   *pixels
@ r1: int       wordcount
@ r2: int       lineno 
@
@ void NEON_doline_n3(uae_u32 *pixels, int wordcount, int lineno);
@
@----------------------------------------------------------------
NEON_doline_n3:
  stmdb     sp!, {r4, lr}

  mov       r3, #1600
  mul       r2, r2, r3
  ldr       r3, =line_data
  add       r2, r3, r2           @ real_bplpt[0]
  add       r3, r2, #200
  add       r4, r3, #200

  @ Load data as early as possible
  vldmia    r4!, {d0}

@ Load masks to registers
  vmov.u8   d10, #0x55
  vmov.u8   d11, #0x33
  vmov.u8   d12, #0x0f

NEON_doline_n3_loop:
@ Load from real_bplpt (now loaded earlier)
@  vld1.8    d0, [r4]!
@  vld1.8    d4, [r2]!
@  vld1.8    d6, [r3]!

  @ Load data as early as possible
  vldmia    r2!, {d4}
  vldmia    r3!, {d6}

@      MERGE_0(b4, b5, 0x55555555, 1);
  vshr.u8   d8, d0, #1		  @ tmp = b >> shift
  vand.8    d2, d8, d10     @ a = tmp & mask
  vand.8    d0, d0, d10     @ b = b & mask

@      MERGE (b6, b7, 0x55555555, 1);
  vshr.u8   d8, d4, #1      @ tmpb = b >> shift
  vshl.u8   d9, d6, #1      @ tmpa = a << shift
  vbit.u8   d6, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d10     @ b = b and bit set from tmpa if mask is false

@      MERGE (b4, b6, 0x33333333, 2);
  vshr.u8   d8, d6, #2      @ tmpb = b >> shift
  vshl.u8   d9, d2, #2      @ tmpa = a << shift
  vbit.u8   d2, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d6, d9, d11     @ b = b and bit set from tmpa if mask is false
@      MERGE (b5, b7, 0x33333333, 2);
  vshr.u8   d8, d4, #2      @ tmpb = b >> shift
  vshl.u8   d9, d0, #2      @ tmpa = a << shift
  vbit.u8   d0, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d11     @ b = b and bit set from tmpa if mask is false

@      MERGE_0(b0, b4, 0x0f0f0f0f, 4);
  vshr.u8   d8, d2, #4		  @ tmp = b >> shift
  vand.8    d3, d8, d12     @ a = tmp & mask
  vand.8    d2, d2, d12     @ b = b & mask
@      MERGE_0(b1, b5, 0x0f0f0f0f, 4);
  vshr.u8   d8, d0, #4			@ tmp = b >> shift
  vand.8    d1, d8, d12     @ a = tmp & mask
  vand.8    d0, d0, d12     @ b = b & mask
@      MERGE_0(b2, b6, 0x0f0f0f0f, 4);
  vshr.u8   d8, d6, #4			@ tmp = b >> shift
  vand.8    d7, d8, d12     @ a = tmp & mask
  vand.8    d6, d6, d12     @ b = b & mask
@      MERGE_0(b3, b7, 0x0f0f0f0f, 4);
  vshr.u8   d8, d4, #4			@ tmp = b >> shift
  vand.8    d5, d8, d12     @ a = tmp & mask
  vand.8    d4, d4, d12     @ b = b & mask
  
  vzip.8    d3, d7
  vzip.8    d1, d5
  vzip.8    d2, d6
  vzip.8    d0, d4

  vzip.8    d3, d1
  vzip.8    d2, d0
  vzip.32   d3, d2
  vzip.32   d1, d0

  vst1.8    {d0, d1, d2, d3}, [r0]!
  
  cmp       r1, #1    @ Exit from here if odd number of words
  ldmeqia   sp!, {r4, pc}
  
  subs      r1, r1, #2    @ We handle 2 words (64 bit) per loop: wordcount -= 2

  @ Load next data (if needed) as early as possible
  vldmiagt  r4!, {d0}

  vzip.8    d7, d5
  vzip.8    d6, d4
  vzip.32   d7, d6
  vzip.32   d5, d4

  vst1.8    {d4, d5, d6, d7}, [r0]!

  bgt       NEON_doline_n3_loop
  
NEON_doline_n3_exit:
  ldmia     sp!, {r4, pc}


.align 8

@----------------------------------------------------------------
@ NEON_doline_n4
@
@ r0: uae_u32   *pixels
@ r1: int       wordcount
@ r2: int       lineno 
@
@ void NEON_doline_n4(uae_u32 *pixels, int wordcount, int lineno);
@
@----------------------------------------------------------------
NEON_doline_n4:
  stmdb     sp!, {r4-r5, lr}

  mov       r3, #1600
  mul       r2, r2, r3
  ldr       r3, =line_data
  add       r2, r3, r2           @ real_bplpt[0]
  add       r3, r2, #200
  add       r4, r3, #200
  add       r5, r4, #200

  @ Load data as early as possible
  vldmia    r4!, {d0}
  vldmia    r5!, {d2}

@ Load masks to registers
  vmov.u8   d10, #0x55
  vmov.u8   d11, #0x33
  vmov.u8   d12, #0x0f

NEON_doline_n4_loop:
@ Load from real_bplpt (now loaded earlier)
@  vld1.8    d0, [r4]!
@  vld1.8    d2, [r5]!
@  vld1.8    d4, [r2]!
@  vld1.8    d6, [r3]!

  @ Load data as early as possible
  vldmia    r2!, {d4}

@      MERGE (b4, b5, 0x55555555, 1);
  vshr.u8   d8, d0, #1      @ tmpb = b >> shift
  vshl.u8   d9, d2, #1      @ tmpa = a << shift

  vldmia    r3!, {d6}

  vbit.u8   d2, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d0, d9, d10     @ b = b and bit set from tmpa if mask is false
@      MERGE (b6, b7, 0x55555555, 1);
  vshr.u8   d8, d4, #1      @ tmpb = b >> shift
  vshl.u8   d9, d6, #1      @ tmpa = a << shift
  vbit.u8   d6, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d10     @ b = b and bit set from tmpa if mask is false

@      MERGE (b4, b6, 0x33333333, 2);
  vshr.u8   d8, d6, #2      @ tmpb = b >> shift
  vshl.u8   d9, d2, #2      @ tmpa = a << shift
  vbit.u8   d2, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d6, d9, d11     @ b = b and bit set from tmpa if mask is false
@      MERGE (b5, b7, 0x33333333, 2);
  vshr.u8   d8, d4, #2      @ tmpb = b >> shift
  vshl.u8   d9, d0, #2      @ tmpa = a << shift
  vbit.u8   d0, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d11     @ b = b and bit set from tmpa if mask is false

@      MERGE_0(b0, b4, 0x0f0f0f0f, 4);
  vshr.u8   d8, d2, #4		  @ tmp = b >> shift
  vand.8    d3, d8, d12     @ a = tmp & mask
  vand.8    d2, d2, d12     @ b = b & mask
@      MERGE_0(b1, b5, 0x0f0f0f0f, 4);
  vshr.u8   d8, d0, #4			@ tmp = b >> shift
  vand.8    d1, d8, d12     @ a = tmp & mask
  vand.8    d0, d0, d12     @ b = b & mask
@      MERGE_0(b2, b6, 0x0f0f0f0f, 4);
  vshr.u8   d8, d6, #4			@ tmp = b >> shift
  vand.8    d7, d8, d12     @ a = tmp & mask
  vand.8    d6, d6, d12     @ b = b & mask
@      MERGE_0(b3, b7, 0x0f0f0f0f, 4);
  vshr.u8   d8, d4, #4			@ tmp = b >> shift
  vand.8    d5, d8, d12     @ a = tmp & mask
  vand.8    d4, d4, d12     @ b = b & mask
  
  vzip.8    d3, d7
  vzip.8    d1, d5
  vzip.8    d2, d6
  vzip.8    d0, d4

  vzip.8    d3, d1
  vzip.8    d2, d0
  vzip.32   d3, d2
  vzip.32   d1, d0

  vst1.8    {d0, d1, d2, d3}, [r0]!
  
  cmp       r1, #1    @ Exit from here if odd number of words
  ldmeqia   sp!, {r4-r5, pc}
  
  subs      r1, r1, #2    @ We handle 2 words (64 bit) per loop: wordcount -= 2

  @ Load next data (if needed) as early as possible
  vldmiagt  r4!, {d0}

  vzip.8    d7, d5
  vzip.8    d6, d4

  vldmiagt  r5!, {d2}

  vzip.32   d7, d6
  vzip.32   d5, d4

  vst1.8    {d4, d5, d6, d7}, [r0]!

  bgt       NEON_doline_n4_loop
  
NEON_doline_n4_exit:
  ldmia     sp!, {r4-r5, pc}


.align 8

@----------------------------------------------------------------
@ NEON_doline_n6
@
@ r0: uae_u32   *pixels
@ r1: int       wordcount
@ r2: int       lineno 
@
@ void NEON_doline_n6(uae_u32 *pixels, int wordcount, int lineno);
@
@----------------------------------------------------------------
NEON_doline_n6:
  stmdb     sp!, {r4-r7, lr}
  
  mov       r3, #1600
  mul       r2, r2, r3
  ldr       r3, =line_data
  add       r2, r3, r2           @ real_bplpt[0]
  add       r3, r2, #200
  add       r4, r3, #200
  add       r5, r4, #200
  add       r6, r5, #200
  add       r7, r6, #200
  
@ Load masks to registers
  vmov.u8   d10, #0x55
  vmov.u8   d11, #0x33
  vmov.u8   d12, #0x0f

NEON_doline_n6_loop:
  @ Load data as early as possible
  vldmia    r6!, {d5}
  vldmia    r7!, {d7}

  @ Load data as early as possible
  vldmia    r4!, {d0}
@      MERGE (b2, b3, 0x55555555, 1);
  vshr.u8   d8, d5, #1      @ tmpb = b >> shift
  vshl.u8   d9, d7, #1      @ tmpa = a << shift
  @ Load data as early as possible
  vldmia    r5!, {d2}
  vbit.u8   d7, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d5, d9, d10     @ b = b and bit set from tmpa if mask is false
  @ Load data as early as possible
  vldmia    r2!, {d4}
@      MERGE (b4, b5, 0x55555555, 1);
  vshr.u8   d8, d0, #1      @ tmpb = b >> shift
  vshl.u8   d9, d2, #1      @ tmpa = a << shift
  @ Load data as early as possible
  vldmia    r3!, {d6}
  vbit.u8   d2, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d0, d9, d10     @ b = b and bit set from tmpa if mask is false
@      MERGE (b6, b7, 0x55555555, 1);
  vshr.u8   d8, d4, #1      @ tmpb = b >> shift
  vshl.u8   d9, d6, #1      @ tmpa = a << shift
  vbit.u8   d6, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d10     @ b = b and bit set from tmpa if mask is false

@      MERGE_0(b0, b2, 0x33333333, 2);
  vshr.u8   d8, d7, #2		  @ tmp = b >> shift
  vand.8    d3, d8, d11     @ a = tmp & mask
  vand.8    d7, d7, d11     @ b = b & mask
@      MERGE_0(b1, b3, 0x33333333, 2);
  vshr.u8   d8, d5, #2		  @ tmp = b >> shift
  vand.8    d1, d8, d11     @ a = tmp & mask
  vand.8    d5, d5, d11     @ b = b & mask
@      MERGE (b4, b6, 0x33333333, 2);
  vshr.u8   d8, d6, #2      @ tmpb = b >> shift
  vshl.u8   d9, d2, #2      @ tmpa = a << shift
  vbit.u8   d2, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d6, d9, d11     @ b = b and bit set from tmpa if mask is false
@      MERGE (b5, b7, 0x33333333, 2);
  vshr.u8   d8, d4, #2      @ tmpb = b >> shift
  vshl.u8   d9, d0, #2      @ tmpa = a << shift
  vbit.u8   d0, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d11     @ b = b and bit set from tmpa if mask is false

@      MERGE (b0, b4, 0x0f0f0f0f, 4);
  vshr.u8   d8, d2, #4      @ tmpb = b >> shift
  vshl.u8   d9, d3, #4      @ tmpa = a << shift
  vbit.u8   d3, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d2, d9, d12     @ b = b and bit set from tmpa if mask is false
@      MERGE (b1, b5, 0x0f0f0f0f, 4);
  vshr.u8   d8, d0, #4      @ tmpb = b >> shift
  vshl.u8   d9, d1, #4      @ tmpa = a << shift
  vbit.u8   d1, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d0, d9, d12     @ b = b and bit set from tmpa if mask is false
@      MERGE (b2, b6, 0x0f0f0f0f, 4);
  vshr.u8   d8, d6, #4      @ tmpb = b >> shift
  vshl.u8   d9, d7, #4      @ tmpa = a << shift
  vbit.u8   d7, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d6, d9, d12     @ b = b and bit set from tmpa if mask is false
@      MERGE (b3, b7, 0x0f0f0f0f, 4);
  vshr.u8   d8, d4, #4      @ tmpb = b >> shift
  vshl.u8   d9, d5, #4      @ tmpa = a << shift
  vbit.u8   d5, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d12     @ b = b and bit set from tmpa if mask is false

  vzip.8    d3, d7
  vzip.8    d1, d5
  vzip.8    d2, d6
  vzip.8    d0, d4

  vzip.8    d3, d1
  vzip.8    d2, d0
  vzip.32   d3, d2
  vzip.32   d1, d0

  vst1.8    {d0, d1, d2, d3}, [r0]!
  
  cmp       r1, #1    @ Exit from here if odd number of words
  ldmeqia   sp!, {r4-r7, pc}

  subs      r1, r1, #2    @ We handle 2 words (64 bit) per loop: wordcount -= 2

  vzip.8    d7, d5
  vzip.8    d6, d4
  vzip.32   d7, d6
  vzip.32   d5, d4

  vst1.8    {d4, d5, d6, d7}, [r0]!

  bgt       NEON_doline_n6_loop
  
NEON_doline_n6_exit:
  ldmia     sp!, {r4-r7, pc}
 

.align 8

@----------------------------------------------------------------
@ NEON_doline_n8
@
@ r0: uae_u32   *pixels
@ r1: int       wordcount
@ r2: int       lineno 
@
@ void NEON_doline_n8(uae_u32 *pixels, int wordcount, int lineno);
@
@----------------------------------------------------------------
NEON_doline_n8:
  stmdb     sp!, {r4-r9, lr}
  
  mov       r3, #1600
  mul       r2, r2, r3
  ldr       r3, =line_data
  add       r2, r3, r2           @ real_bplpt[0]
  add       r3, r2, #200
  add       r4, r3, #200
  add       r5, r4, #200
  add       r6, r5, #200
  add       r7, r6, #200
  add       r8, r7, #200
  add       r9, r8, #200
  
  @ Load data as early as possible
  vldmia    r8!, {d1}
  vldmia    r9!, {d3}

@ Load masks to registers
  vmov.u8   d10, #0x55
  vmov.u8   d11, #0x33
  vmov.u8   d12, #0x0f

NEON_doline_n8_loop:
  @ Load data as early as possible
  vldmia    r6!, {d5}
@      MERGE (b0, b1, 0x55555555, 1);
  vshr.u8   d8, d1, #1      @ tmpb = b >> shift
  vshl.u8   d9, d3, #1      @ tmpa = a << shift
  @ Load data as early as possible
  vldmia    r7!, {d7}
  vbit.u8   d3, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d1, d9, d10     @ b = b and bit set from tmpa if mask is false
  @ Load data as early as possible
  vldmia    r4!, {d0}
@      MERGE (b2, b3, 0x55555555, 1);
  vshr.u8   d8, d5, #1      @ tmpb = b >> shift
  vshl.u8   d9, d7, #1      @ tmpa = a << shift
  @ Load data as early as possible
  vldmia    r5!, {d2}
  vbit.u8   d7, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d5, d9, d10     @ b = b and bit set from tmpa if mask is false
  @ Load data as early as possible
  vldmia    r2!, {d4}
@      MERGE (b4, b5, 0x55555555, 1);
  vshr.u8   d8, d0, #1      @ tmpb = b >> shift
  vshl.u8   d9, d2, #1      @ tmpa = a << shift
  @ Load data as early as possible
  vldmia    r3!, {d6}
  vbit.u8   d2, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d0, d9, d10     @ b = b and bit set from tmpa if mask is false
@      MERGE (b6, b7, 0x55555555, 1);
  vshr.u8   d8, d4, #1      @ tmpb = b >> shift
  vshl.u8   d9, d6, #1      @ tmpa = a << shift
  vbit.u8   d6, d8, d10     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d10     @ b = b and bit set from tmpa if mask is false

@      MERGE (b0, b2, 0x33333333, 2);
  vshr.u8   d8, d7, #2      @ tmpb = b >> shift
  vshl.u8   d9, d3, #2      @ tmpa = a << shift
  vbit.u8   d3, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d7, d9, d11     @ b = b and bit set from tmpa if mask is false
@      MERGE (b1, b3, 0x33333333, 2);
  vshr.u8   d8, d5, #2      @ tmpb = b >> shift
  vshl.u8   d9, d1, #2      @ tmpa = a << shift
  vbit.u8   d1, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d5, d9, d11     @ b = b and bit set from tmpa if mask is false
@      MERGE (b4, b6, 0x33333333, 2);
  vshr.u8   d8, d6, #2      @ tmpb = b >> shift
  vshl.u8   d9, d2, #2      @ tmpa = a << shift
  vbit.u8   d2, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d6, d9, d11     @ b = b and bit set from tmpa if mask is false
@      MERGE (b5, b7, 0x33333333, 2);
  vshr.u8   d8, d4, #2      @ tmpb = b >> shift
  vshl.u8   d9, d0, #2      @ tmpa = a << shift
  vbit.u8   d0, d8, d11     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d11     @ b = b and bit set from tmpa if mask is false

@      MERGE (b0, b4, 0x0f0f0f0f, 4);
  vshr.u8   d8, d2, #4      @ tmpb = b >> shift
  vshl.u8   d9, d3, #4      @ tmpa = a << shift
  vbit.u8   d3, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d2, d9, d12     @ b = b and bit set from tmpa if mask is false
@      MERGE (b1, b5, 0x0f0f0f0f, 4);
  vshr.u8   d8, d0, #4      @ tmpb = b >> shift
  vshl.u8   d9, d1, #4      @ tmpa = a << shift
  vbit.u8   d1, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d0, d9, d12     @ b = b and bit set from tmpa if mask is false
@      MERGE (b2, b6, 0x0f0f0f0f, 4);
  vshr.u8   d8, d6, #4      @ tmpb = b >> shift
  vshl.u8   d9, d7, #4      @ tmpa = a << shift
  vbit.u8   d7, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d6, d9, d12     @ b = b and bit set from tmpa if mask is false
@      MERGE (b3, b7, 0x0f0f0f0f, 4);
  vshr.u8   d8, d4, #4      @ tmpb = b >> shift
  vshl.u8   d9, d5, #4      @ tmpa = a << shift
  vbit.u8   d5, d8, d12     @ a = a and bit set from tmpb if mask is true 
  vbif.u8   d4, d9, d12     @ b = b and bit set from tmpa if mask is false

  vzip.8    d3, d7
  vzip.8    d1, d5
  vzip.8    d2, d6
  vzip.8    d0, d4

  vzip.8    d3, d1
  vzip.8    d2, d0
  vzip.32   d3, d2
  vzip.32   d1, d0

  vst1.8    {d0, d1, d2, d3}, [r0]!
  
  cmp       r1, #1    @ Exit from here if odd number of words
  ldmeqia   sp!, {r4-r9, pc}

  subs      r1, r1, #2    @ We handle 2 words (64 bit) per loop: wordcount -= 2

  @ Load data as early as possible
  vldmiagt  r8!, {d1}
  
  vzip.8    d7, d5
  vzip.8    d6, d4

  @ Load data as early as possible
  vldmiagt  r9!, {d3}

  vzip.32   d7, d6
  vzip.32   d5, d4

  vst1.8    {d4, d5, d6, d7}, [r0]!

  bgt       NEON_doline_n8_loop
  
NEON_doline_n8_exit:
  ldmia     sp!, {r4-r9, pc}


.data

.align 8

Lookup_doline_n1:
  .long 0x00000000, 0x01000000, 0x00010000, 0x01010000
  .long 0x00000100, 0x01000100, 0x00010100, 0x01010100
  .long 0x00000001, 0x01000001, 0x00010001, 0x01010001
  .long 0x00000101, 0x01000101, 0x00010101, 0x01010101
